float4x4 World; 
float4x4 View; 
float4x4 Projection; 

float4 vecLightDir[3];
float4 LightColor[3];
float4 vecEye;

float NumLights;

float FarClip = 40;

texture Texture;

bool SelfShadow = true;

float3 ModelPos;

sampler Sampler = sampler_state
{
    Texture = (Texture);

    MinFilter = Linear;
    MagFilter = Linear;
    MipFilter = Linear;
};

texture ShadowMap;

sampler shadowSampler = sampler_state
{
    Texture = <ShadowMap>;
    
    MinFilter = LINEAR;
    MagFilter = LINEAR;
    MipFilter = LINEAR;
    
    AddressU = CLAMP;
    AddressV = CLAMP;
};

float3 LightPosition;
float3 CameraPosition;

float4x4 WorldLightViewProjection;

// Vertex shader input structure.
struct VS_INPUT
{
    float4 Position : POSITION0;
    float3 Normal : NORMAL0;
    float2 TexCoord : TEXCOORD0;
};


// Vertex shader output structure.
struct VS_OUTPUT
{
	float4 Position : POSITION0;
    float2 TexCoord : TEXCOORD0;
    float3 Normal : TEXCOORD1;
    float3 View : TEXCOORD2;
    float4 WorldPosition : TEXCOORD3;
    float4 ShadowProjection : TEXCOORD4;
};


// Vertex shader program.
VS_OUTPUT VertexShader(VS_INPUT input)
{
    VS_OUTPUT output;
    
    // Skin the vertex position.
    float4 position = mul(input.Position, World);
    
    output.Position = mul(mul(position, View), Projection);

    // Skin the vertex normal, then compute lighting.
    float3 normal = normalize(mul(input.Normal, World));
    output.Normal = normal;

    output.TexCoord = input.TexCoord;
    
    output.View = vecEye - position;
    
    output.WorldPosition = mul(input.Position, World);
    output.WorldPosition /= output.WorldPosition.w;

	output.ShadowProjection = mul(input.Position, WorldLightViewProjection);
    
    return output;
}


// Pixel shader program.
float4 PixelShader(VS_OUTPUT input) : COLOR0
{
	float4 Color = float4 (0, 0, 0, 1);
	
	for (float i = 0; i < NumLights; i++)
	{
		float4 diffuse = {LightColor[i].rgb, 1.0f};
		float4 ambient = { 0.1f, 0.1f, 0.1f, 1.0f};

		float3 Normal = normalize(input.Normal);
		float3 LightDir = normalize(vecLightDir[i]);
		float3 ViewDir = normalize(input.View); 
    
		float4 diff = saturate(dot(Normal, LightDir)); // diffuse component

		// R = 2 * (N.L) * N - L
		float3 Reflect = normalize(2 * diff * Normal - LightDir); 
		float4 specular = pow(saturate(dot(Reflect, ViewDir)), LightColor[i].a); // R.V^n

		// I = Acolor + Dcolor * N.L + (R.V)n
		Color += ambient + diffuse  * diff + specular; 
	}
	
	float len = length(LightPosition - input.WorldPosition) / FarClip;

	float4 tex = tex2D(Sampler, input.TexCoord);
	float4 Final = tex * Color;
	
	float2 ProjectedTexCoords;


	ProjectedTexCoords[0] = input.ShadowProjection.x / 
                                input.ShadowProjection.w / 2.0f + 0.5f;
    ProjectedTexCoords[1] = -input.ShadowProjection.y / 
                                input.ShadowProjection.w / 2.0f + 0.5f;


	float4 moments = tex2D(shadowSampler, ProjectedTexCoords);
    
    float E_x2 = moments.y;
    float Ex_2 = moments.x * moments.x;
    float variance = min(max(E_x2 - Ex_2, 0.0) + 0.0001f, 1.0);
    float m_d = (moments.x - len);
    float p = variance / (variance + m_d * m_d);
        
    float shadow = max(step(len, moments.x), p);

	shadow = clamp(shadow + .4, 0, 1);

	if (SelfShadow && distance(input.WorldPosition.xz, ModelPos.xz) < 8)
		Final *= shadow;
	
	
	Final.a = tex.a;
	
	return Final;
}


technique Model
{
    pass Pass1
    {	
        VertexShader = compile vs_1_1 VertexShader();
        PixelShader = compile ps_2_b PixelShader();
    }
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////////
///// DEPTH /////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct DEPTH_VS_INPUT
{
    float4 Position         : POSITION;
    float2 TexCoords        : TEXCOORD0;
};


struct DEPTH_VS_OUTPUT
{
    float4 Position         : POSITION;
    float2 TexCoords        : TEXCOORD0;
    float4 WorldPosition    : TEXCOORD1;
};

DEPTH_VS_OUTPUT Depth_VS (DEPTH_VS_INPUT Input)
{
    DEPTH_VS_OUTPUT Output;
    
    float4x4 WorldCameraViewProjection = mul(World, mul(View, Projection));
    
    Output.Position = mul(float4(Input.Position.xyz, 1.0f), WorldCameraViewProjection);

    Output.WorldPosition = mul(Input.Position, World);
    Output.WorldPosition /= Output.WorldPosition.w;
    
    Output.TexCoords = Input.TexCoords;
    
    return Output;
}

float4 Depth_PS (DEPTH_VS_OUTPUT Input) : COLOR0
{
    float depth = length(LightPosition - Input.WorldPosition) / FarClip;
    
    return float4(depth, depth * depth, 0, 1);
}

technique Depth
{
    pass P0
    {
        VertexShader = compile vs_2_0 Depth_VS();
        PixelShader  = compile ps_2_0 Depth_PS();
    }
}
